@@ -1,53 +1,63 @@
+/*
+ * Copyright 2011, Ben Langmead <langmea@cs.jhu.edu>
+ *
+ * This file is part of Bowtie 2.
+ *
+ * Bowtie 2 is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
+ *
+ * Bowtie 2 is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with Bowtie 2.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
 #ifndef BLOCKWISE_SA_H_
 #define BLOCKWISE_SA_H_
 
-#if (__cplusplus >= 201103)
-#include <thread>
-#endif
-
 #include <stdint.h>
 #include <stdlib.h>
 #include <iostream>
 #include <sstream>
 #include <stdexcept>
-
-#include "alphabet.h"
 #include "assert_helpers.h"
-#include "binary_sa_search.h"
 #include "diff_sample.h"
-#include "ds.h"
 #include "multikey_qsort.h"
 #include "random_source.h"
-#include "threading.h"
-#include "timer.h"
-#include "util.h"
-#include "word_io.h"
+#include "binary_sa_search.h"
 #include "zbox.h"
+#include "alphabet.h"
+#include "timer.h"
+#include "ds.h"
+#include "mem_ids.h"
 
 using namespace std;
 
 // Helpers for printing verbose messages
 
 #ifndef VMSG_NL
-#define VMSG_NL(...)				\
-	if(this->verbose()) {			\
-		stringstream tmp;		\
-		tmp << __VA_ARGS__ << endl;	\
-		this->verbose(tmp.str());	\
-	}
+#define VMSG_NL(...) \
+if(this->verbose()) { \
+	stringstream tmp; \
+	tmp << __VA_ARGS__ << endl; \
+	this->verbose(tmp.str()); \
+}
 #endif
 
 #ifndef VMSG
-#define VMSG(...)				\
-	if(this->verbose()) {			\
-		stringstream tmp;		\
-		tmp << __VA_ARGS__;		\
-		this->verbose(tmp.str());	\
-	}
+#define VMSG(...) \
+if(this->verbose()) { \
+	stringstream tmp; \
+	tmp << __VA_ARGS__; \
+	this->verbose(tmp.str()); \
+}
 #endif
 
-#define EBWTB_CAT ((int) 2)
-
 /**
  * Abstract parent class for blockwise suffix-array building schemes.
  */
@@ -55,32 +65,44 @@ template<typename TStr>
 class BlockwiseSA {
 public:
 	BlockwiseSA(const TStr& __text,
-		    TIndexOffU __bucketSz,
-		    bool __sanityCheck = false,
-		    bool __passMemExc = false,
-		    bool __verbose = false,
-		    ostream& __logger = cout) :
-		_text(__text),
-		_bucketSz(max<TIndexOffU>(__bucketSz, 2u)),
-		_sanityCheck(__sanityCheck),
-		_passMemExc(__passMemExc),
-		_verbose(__verbose),
-		_itrBucket(),
-		_itrBucketPos(OFF_MASK),
-		_itrPushedBackSuffix(OFF_MASK),
-		_logger(__logger)
-		{ }
+	            TIndexOffU __bucketSz,
+	            bool __sanityCheck = false,
+	            bool __passMemExc = false,
+	            bool __verbose = false,
+	            ostream& __logger = cout) :
+	_text(__text),
+	_bucketSz(max<TIndexOffU>(__bucketSz, 2u)),
+	_sanityCheck(__sanityCheck),
+	_passMemExc(__passMemExc),
+	_verbose(__verbose),
+	_itrBucket(EBWTB_CAT),
+	_itrBucketPos(OFF_MASK),
+	_itrPushedBackSuffix(OFF_MASK),
+	_logger(__logger)
+	{ }
 
-	virtual ~BlockwiseSA()
-#if __cplusplus > 199711L
-	noexcept(false)
-#endif
-		{ }
+	virtual ~BlockwiseSA() { }
 
 	/**
 	 * Get the next suffix; compute the next bucket if necessary.
 	 */
-	virtual TIndexOffU nextSuffix() = 0;
+	TIndexOffU nextSuffix() {
+		if(_itrPushedBackSuffix != OFF_MASK) {
+			TIndexOffU tmp = _itrPushedBackSuffix;
+			_itrPushedBackSuffix = OFF_MASK;
+			return tmp;
+		}
+		while(_itrBucketPos >= _itrBucket.size() ||
+		      _itrBucket.size() == 0)
+		{
+			if(!hasMoreBlocks()) {
+				throw out_of_range("No more suffixes");
+			}
+			nextBlock();
+			_itrBucketPos = 0;
+		}
+		return _itrBucket[_itrBucketPos++];
+	}
 
 	/**
 	 * Return true iff the next call to nextSuffix will succeed.
@@ -113,10 +135,10 @@ public:
 	 * lexicographically-first suffix.
 	 */
 	bool suffixItrIsReset() {
-		return _itrBucket.size()   == 0 &&
-			_itrBucketPos        == OFF_MASK &&
-			_itrPushedBackSuffix == OFF_MASK &&
-			isReset();
+		return _itrBucket.size()    == 0 &&
+		       _itrBucketPos        == OFF_MASK &&
+		       _itrPushedBackSuffix == OFF_MASK &&
+		       isReset();
 	}
 
 	const TStr& text()  const { return _text; }
@@ -136,7 +158,7 @@ protected:
 	 * Grab the next block of sorted suffixes.  The block is guaranteed
 	 * to have at most _bucketSz elements.
 	 */
-	virtual void nextBlock(int cur_block, int tid = 0) = 0;
+	virtual void nextBlock() = 0;
 	/// Return true iff more blocks are available
 	virtual bool hasMoreBlocks() const = 0;
 	/// Optionally output a verbose message
@@ -152,7 +174,7 @@ protected:
 	const bool       _sanityCheck; /// whether to perform sanity checks
 	const bool       _passMemExc;  /// true -> pass on memory exceptions
 	const bool       _verbose;     /// be talkative
-	EList<TIndexOffU> _itrBucket;   /// current bucket
+	EList<TIndexOffU>  _itrBucket;   /// current bucket
 	TIndexOffU         _itrBucketPos;/// offset into current bucket
 	TIndexOffU         _itrPushedBackSuffix; /// temporary slot for lookahead
 	ostream&         _logger;      /// write log messages here
@@ -166,13 +188,13 @@ template<typename TStr>
 class InorderBlockwiseSA : public BlockwiseSA<TStr> {
 public:
 	InorderBlockwiseSA(const TStr& __text,
-			   TIndexOffU __bucketSz,
-			   bool __sanityCheck = false,
-			   bool __passMemExc = false,
-			   bool __verbose = false,
-			   ostream& __logger = cout) :
-		BlockwiseSA<TStr>(__text, __bucketSz, __sanityCheck, __passMemExc, __verbose, __logger)
-		{ }
+	                   TIndexOffU __bucketSz,
+	                   bool __sanityCheck = false,
+	   	               bool __passMemExc = false,
+	                   bool __verbose = false,
+	                   ostream& __logger = cout) :
+	BlockwiseSA<TStr>(__text, __bucketSz, __sanityCheck, __passMemExc, __verbose, __logger)
+	{ }
 };
 
 /**
@@ -185,36 +207,18 @@ public:
 	typedef DifferenceCoverSample<TStr> TDC;
 
 	KarkkainenBlockwiseSA(const TStr& __text,
-			      TIndexOffU __bucketSz,
-			      int __nthreads,
-			      uint32_t __dcV,
-			      uint32_t __seed = 0,
-			      bool __sanityCheck = false,
-			      bool __passMemExc = false,
-			      bool __verbose = false,
-			      string base_fname = "",
-			      ostream& __logger = cout) :
-		InorderBlockwiseSA<TStr>(__text, __bucketSz, __sanityCheck, __passMemExc, __verbose, __logger),
-		_sampleSuffs(), _nthreads(__nthreads), _itrBucketIdx(0), _cur(0), _dcV(__dcV), _dc(NULL), _built(false), _base_fname(base_fname), _bigEndian(currentlyBigEndian()), _done(NULL)
-		{ _randomSrc.init(__seed); reset(); }
+	                      TIndexOffU __bucketSz,
+	                      uint32_t __dcV,
+	                      uint32_t __seed = 0,
+	      	              bool __sanityCheck = false,
+	   	                  bool __passMemExc = false,
+	      	              bool __verbose = false,
+	      	              ostream& __logger = cout) :
+	InorderBlockwiseSA<TStr>(__text, __bucketSz, __sanityCheck, __passMemExc, __verbose, __logger),
+	_sampleSuffs(EBWTB_CAT), _cur(0), _dcV(__dcV), _dc(EBWTB_CAT), _built(false)
+	{ _randomSrc.init(__seed); reset(); }
 
-	~KarkkainenBlockwiseSA()
-#if __cplusplus > 199711L
-	noexcept(false)
-#endif
-		{
-			if(_dc != NULL) delete _dc; // difference cover sample
-			if (_done != NULL) {
-				delete[] _done;
-				_done = NULL;
-			}
-			if(_threads.size() > 0) {
-				for (size_t tid = 0; tid < _threads.size(); tid++) {
-					_threads[tid]->join();
-					delete _threads[tid];
-				}
-			}
-		}
+	~KarkkainenBlockwiseSA() { }
 
 	/**
 	 * Allocate an amount of memory that simulates the peak memory
@@ -227,119 +231,19 @@ public:
 		// _sampleSuffs and _itrBucket are in memory at the peak
 		size_t bsz = bucketSz;
 		size_t sssz = len / max<TIndexOffU>(bucketSz-1, 1);
-		AutoArray<TIndexOffU> tmp(bsz + sssz + (1024 * 1024 /*out of caution*/));
+		AutoArray<TIndexOffU> tmp(bsz + sssz + (1024 * 1024 /*out of caution*/), EBWT_CAT);
 		return bsz;
 	}
 
-
-	static void nextBlock_Worker(void *vp) {
-
-		pair<KarkkainenBlockwiseSA*, int> param = *(pair<KarkkainenBlockwiseSA*, int>*)vp;
-		KarkkainenBlockwiseSA* sa = param.first;
-		int tid = param.second;
-		while(true) {
-			size_t cur = 0;
-			{
-				ThreadSafe ts(&sa->_mutex);
-				cur = sa->_cur;
-				if(cur > sa->_sampleSuffs.size()) break;
-				sa->_cur++;
-			}
-			sa->nextBlock((int)cur, tid);
-			// Write suffixes into a file
-			std::ostringstream number; number << cur;
-			const string fname = sa->_base_fname + "." + number.str() + ".sa";
-			ofstream sa_file(fname.c_str(), ios::binary);
-			if(!sa_file.good()) {
-				cerr << "Could not open file for writing a reference graph: \"" << fname << "\"" << endl;
-				throw 1;
-			}
-			const EList<TIndexOffU>& bucket = sa->_itrBuckets[tid];
-			writeU<TIndexOffU>(sa_file, (TIndexOffU)bucket.size(), sa->_bigEndian);
-			for(size_t i = 0; i < bucket.size(); i++) {
-				writeU<TIndexOffU>(sa_file, bucket[i], sa->_bigEndian);
-			}
-			sa_file.close();
-			sa->_itrBuckets[tid].clear();
-			sa->_done[cur] = true;
-		}
-	}
-
-
-	/**
-	 * Get the next suffix; compute the next bucket if necessary.
-	 */
-	virtual TIndexOffU nextSuffix() {
-		// Launch threads if not
-		if(this->_nthreads > 1) {
-			if(_threads.size() == 0) {
-				_done = new volatile bool[_sampleSuffs.size()+1];
-				for (size_t i = 0; i < _sampleSuffs.size() + 1; i++) {
-					_done[i] = false;
-				}
-				_itrBuckets.resize(this->_nthreads);
-				_tparams.resize(this->_nthreads);
-				for(int tid = 0; tid < this->_nthreads; tid++) {
-					_tparams[tid].first = this;
-					_tparams[tid].second = tid;
-#if (__cplusplus >= 201103L)
-					_threads.push_back(new std::thread(nextBlock_Worker, (void*)&_tparams[tid]));
-#else
-					_threads.push_back(new tthread::thread(nextBlock_Worker, (void*)&_tparams[tid]));
-#endif
-				}
-				assert_eq(_threads.size(), (size_t)this->_nthreads);
-			}
-		}
-		if(this->_itrPushedBackSuffix != OFF_MASK) {
-			TIndexOffU tmp = this->_itrPushedBackSuffix;
-			this->_itrPushedBackSuffix = OFF_MASK;
-			return tmp;
-		}
-		while(this->_itrBucketPos >= this->_itrBucket.size() ||
-		      this->_itrBucket.size() == 0)
-		{
-			if(!hasMoreBlocks()) {
-				throw out_of_range("No more suffixes");
-			}
-			if(this->_nthreads == 1) {
-				nextBlock((int)_cur);
-				_cur++;
-			} else {
-				while(!_done[this->_itrBucketIdx]) {
-					SLEEP(1);
-				}
-				// Read suffixes from a file
-				std::ostringstream number; number << this->_itrBucketIdx;
-				const string fname = _base_fname + "." + number.str() + ".sa";
-				ifstream sa_file(fname.c_str(), ios::binary);
-				if(!sa_file.good()) {
-					cerr << "Could not open file for reading a reference graph: \"" << fname << "\"" << endl;
-					throw 1;
-				}
-				size_t numSAs = readU<TIndexOffU>(sa_file, false /* don't endian swap */);
-				this->_itrBucket.resize(numSAs);
-				for(size_t i = 0; i < numSAs; i++) {
-					this->_itrBucket[i] = readU<TIndexOffU>(sa_file, false);
-				}
-				sa_file.close();
-				std::remove(fname.c_str());
-			}
-			this->_itrBucketIdx++;
-			this->_itrBucketPos = 0;
-		}
-		return this->_itrBucket[this->_itrBucketPos++];
-	}
-
 	/// Defined in blockwise_sa.cpp
-	virtual void nextBlock(int cur_block, int tid = 0);
+	virtual void nextBlock();
 
 	/// Defined in blockwise_sa.cpp
 	virtual void qsort(EList<TIndexOffU>& bucket);
 
 	/// Return true iff more blocks are available
 	virtual bool hasMoreBlocks() const {
-		return this->_itrBucketIdx <= _sampleSuffs.size();
+		return _cur <= _sampleSuffs.size();
 	}
 
 	/// Return the difference-cover period
@@ -373,10 +277,10 @@ private:
 	 */
 	void build() {
 		// Calculate difference-cover sample
-		assert(_dc == NULL);
+		assert(_dc.get() == NULL);
 		if(_dcV != 0) {
-			_dc = new TDC(this->text(), _dcV, this->verbose(), this->sanityCheck());
-			_dc->build(this->_nthreads);
+			_dc.init(new TDC(this->text(), _dcV, this->verbose(), this->sanityCheck()));
+			_dc.get()->build();
 		}
 		// Calculate sample suffixes
 		if(this->bucketSz() <= this->text().length()) {
@@ -384,8 +288,8 @@ private:
 			buildSamples();
 		} else {
 			VMSG_NL("Skipping building samples since text length " <<
-				this->text().length() << " is less than bucket size: " <<
-				this->bucketSz());
+			        this->text().length() << " is less than bucket size: " <<
+			        this->bucketSz());
 		}
 		_built = true;
 	}
@@ -398,66 +302,54 @@ private:
 	 * Defined in blockwise_sa.cpp
 	 */
 	inline bool tieBreakingLcp(TIndexOffU aOff,
-				   TIndexOffU bOff,
-				   TIndexOffU& lcp,
-				   bool& lcpIsSoft);
+	                           TIndexOffU bOff,
+	                           TIndexOffU& lcp,
+	                           bool& lcpIsSoft);
 
 	/**
 	 * Compare two suffixes using the difference-cover sample.
 	 */
 	inline bool suffixCmp(TIndexOffU cmp,
-			      TIndexOffU i,
-			      int64_t& j,
-			      int64_t& k,
-			      bool& kSoft,
-			      const EList<TIndexOffU>& z);
+	                      TIndexOffU i,
+	                      int64_t& j,
+	                      int64_t& k,
+	                      bool& kSoft,
+	                      const EList<TIndexOffU>& z);
 
 	void buildSamples();
 
-	EList<TIndexOffU> _sampleSuffs; /// sample suffixes
-	int                _nthreads; /// # of threads
-	TIndexOffU         _itrBucketIdx;
+	EList<TIndexOffU>  _sampleSuffs; /// sample suffixes
 	TIndexOffU         _cur;         /// offset to 1st elt of next block
 	const uint32_t   _dcV;         /// difference-cover periodicity
-	TDC*             _dc;          /// queryable difference-cover data
+	PtrWrap<TDC>     _dc;          /// queryable difference-cover data
 	bool             _built;       /// whether samples/DC have been built
 	RandomSource     _randomSrc;   /// source of pseudo-randoms
-	MUTEX_T                 _mutex;       /// synchronization of output message
-	string                  _base_fname;  /// base file name for storing SA blocks
-	bool                    _bigEndian;   /// bigEndian?
-#if (__cplusplus >= 201103L)
-	EList<std::thread*> _threads;    /// thread list
-#else
-	EList<tthread::thread*> _threads;     /// thread list
-#endif
-	EList<pair<KarkkainenBlockwiseSA*, int> > _tparams;
-	EList<EList<TIndexOffU> >     _itrBuckets;  /// buckets
-	volatile bool* _done;        /// is a block processed?
 };
 
 /**
  * Qsort the set of suffixes whose offsets are in 'bucket'.
  */
 template<typename TStr>
-void KarkkainenBlockwiseSA<TStr>::qsort(EList<TIndexOffU>& bucket) {
+inline void KarkkainenBlockwiseSA<TStr>::qsort(EList<TIndexOffU>& bucket) {
 	const TStr& t = this->text();
-	TIndexOffU *s = &bucket[0];
-	TIndexOffU slen = (TIndexOffU)bucket.size();
+	TIndexOffU *s = bucket.ptr();
+	size_t slen = bucket.size();
 	TIndexOffU len = (TIndexOffU)t.length();
-	if(_dc != NULL) {
+	if(_dc.get() != NULL) {
 		// Use the difference cover as a tie-breaker if we have it
 		VMSG_NL("  (Using difference cover)");
 		// Extract the 'host' array because it's faster to work
-		// with than the String<> container
-		uint8_t *host = (uint8_t*)&t[0];
-		mkeyQSortSufDcU8(t, host, len, s, slen, *_dc, 4,
-				 this->verbose(), this->sanityCheck());
+		// with than the EList<> container
+		const uint8_t *host = (const uint8_t *)t.buf();
+		assert(_dc.get() != NULL);
+		mkeyQSortSufDcU8(t, host, len, s, slen, *_dc.get(), 4,
+		                 this->verbose(), this->sanityCheck());
 	} else {
 		VMSG_NL("  (Not using difference cover)");
 		// We don't have a difference cover - just do a normal
 		// suffix sort
 		mkeyQSortSuf(t, s, slen, 4,
-			     this->verbose(), this->sanityCheck());
+		             this->verbose(), this->sanityCheck());
 	}
 }
 
@@ -468,65 +360,26 @@ void KarkkainenBlockwiseSA<TStr>::qsort(EList<TIndexOffU>& bucket) {
  * packed means that the array cannot be sorted directly.
  */
 template<>
-void KarkkainenBlockwiseSA<S2bDnaString>::qsort(EList<TIndexOffU>& bucket) {
+inline void KarkkainenBlockwiseSA<S2bDnaString>::qsort(
+	EList<TIndexOffU>& bucket)
+{
 	const S2bDnaString& t = this->text();
-	TIndexOffU *s = &bucket[0];
-	TIndexOffU slen = (TIndexOffU)bucket.size();
-	TIndexOffU len = (TIndexOffU)t.length();
-	if(_dc != NULL) {
+	TIndexOffU *s = bucket.ptr();
+	size_t slen = bucket.size();
+	size_t len = t.length();
+	if(_dc.get() != NULL) {
 		// Use the difference cover as a tie-breaker if we have it
 		VMSG_NL("  (Using difference cover)");
 		// Can't use the text's 'host' array because the backing
 		// store for the packed string is not one-char-per-elt.
-		mkeyQSortSufDcU8(t, t, len, s, slen, *_dc, 4,
-				 this->verbose(), this->sanityCheck());
+		mkeyQSortSufDcU8(t, t, len, s, slen, *_dc.get(), 4,
+		                 this->verbose(), this->sanityCheck());
 	} else {
 		VMSG_NL("  (Not using difference cover)");
 		// We don't have a difference cover - just do a normal
 		// suffix sort
 		mkeyQSortSuf(t, s, slen, 4,
-			     this->verbose(), this->sanityCheck());
-	}
-}
-
-template<typename TStr>
-struct BinarySortingParam {
-	const TStr*              t;
-	const EList<TIndexOffU>* sampleSuffs;
-	EList<TIndexOffU>        bucketSzs;
-	EList<TIndexOffU>        bucketReps;
-	size_t                   begin;
-	size_t                   end;
-};
-
-template<typename TStr>
-static void BinarySorting_worker(void *vp)
-{
-	BinarySortingParam<TStr>* param = (BinarySortingParam<TStr>*)vp;
-	const TStr& t = *(param->t);
-	size_t len = t.length();
-	const EList<TIndexOffU>& sampleSuffs = *(param->sampleSuffs);
-	EList<TIndexOffU>& bucketSzs = param->bucketSzs;
-	EList<TIndexOffU>& bucketReps = param->bucketReps;
-	ASSERT_ONLY(size_t numBuckets = bucketSzs.size());
-	size_t begin = param->begin;
-	size_t end = param->end;
-	// Iterate through every suffix in the text, determine which
-	// bucket it falls into by doing a binary search across the
-	// sorted list of samples, and increment a counter associated
-	// with that bucket.  Also, keep one representative for each
-	// bucket so that we can split it later.  We loop in ten
-	// stretches so that we can print out a helpful progress
-	// message.  (This step can take a long time.)
-	for(TIndexOffU i = (TIndexOffU)begin; i < end && i < len; i++) {
-		TIndexOffU r = binarySASearch(t, i, sampleSuffs);
-		if(r == std::numeric_limits<TIndexOffU>::max()) continue; // r was one of the samples
-		assert_lt(r, numBuckets);
-		bucketSzs[r]++;
-		assert_lt(bucketSzs[r], len);
-		if(bucketReps[r] == OFF_MASK || (i & 100) == 0) {
-			bucketReps[r] = i; // clobbers previous one, but that's OK
-		}
+		             this->verbose(), this->sanityCheck());
 	}
 }
 
@@ -547,28 +400,27 @@ void KarkkainenBlockwiseSA<TStr>::buildSamples() {
 	assert_gt(numSamples, 0);
 	VMSG_NL("Reserving space for " << numSamples << " sample suffixes");
 	if(this->_passMemExc) {
-		// reserve(_sampleSuffs, numSamples, Exact());
-		_sampleSuffs.resize(numSamples);
+		_sampleSuffs.resizeExact(numSamples);
 		// Randomly generate samples.  Allow duplicates for now.
 		VMSG_NL("Generating random suffixes");
 		for(size_t i = 0; i < numSamples; i++) {
-#ifdef BOWTIE_64BIT_INDEX
-			_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU64() % len);
+#ifdef BOWTIE_64BIT_INDEX         
+			_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU64() % len); 
 #else
-			_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU32() % len);
+			_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU32() % len); 
 #endif
 		}
 	} else {
 		try {
-			_sampleSuffs.resize(numSamples);
+			_sampleSuffs.resizeExact(numSamples);
 			// Randomly generate samples.  Allow duplicates for now.
 			VMSG_NL("Generating random suffixes");
 			for(size_t i = 0; i < numSamples; i++) {
 #ifdef BOWTIE_64BIT_INDEX
-				_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU64() % len);
+				_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU64() % len); 
 #else
-				_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU32() % len);
-#endif
+				_sampleSuffs[i] = (TIndexOffU)(_randomSrc.nextU32() % len); 
+#endif                
 			}
 		} catch(bad_alloc &e) {
 			if(this->_passMemExc) {
@@ -591,86 +443,75 @@ void KarkkainenBlockwiseSA<TStr>::buildSamples() {
 		VMSG_NL("QSorting " << _sampleSuffs.size() << " sample offsets, eliminating duplicates");
 		_sampleSuffs.sort();
 		size_t sslen = _sampleSuffs.size();
-		for (size_t i = 0; i < sslen-1; i++) {
-			if (_sampleSuffs[i] == _sampleSuffs[i+1]) {
+		for(size_t i = 0; i < sslen-1; i++) {
+			if(_sampleSuffs[i] == _sampleSuffs[i+1]) {
 				_sampleSuffs.erase(i--);
 				sslen--;
 			}
 		}
-		// Multikey quicksort the samples
-		{
-			Timer timer(cout, "  Multikey QSorting samples time: ", this->verbose());
-			VMSG_NL("Multikey QSorting " << _sampleSuffs.size() << " samples");
-			this->qsort(_sampleSuffs);
-		}
+	}
+	// Multikey quicksort the samples
+	{
+		Timer timer(cout, "  Multikey QSorting samples time: ", this->verbose());
+		VMSG_NL("Multikey QSorting " << _sampleSuffs.size() << " samples");
+		this->qsort(_sampleSuffs);
 	}
 	// Calculate bucket sizes
 	VMSG_NL("Calculating bucket sizes");
 	int limit = 5;
 	// Iterate until all buckets are less than
 	while(--limit >= 0) {
+		// Calculate bucket sizes by doing a binary search for each
+		// suffix and noting where it lands
 		TIndexOffU numBuckets = (TIndexOffU)_sampleSuffs.size()+1;
-#if (__cplusplus >= 201103l)
-		AutoArray<std::thread*> threads(this->_nthreads);
-#else
-		AutoArray<tthread::thread*> threads(this->_nthreads);
-#endif
-		EList<BinarySortingParam<TStr> > tparams;
-		tparams.resize(this->_nthreads);
-		for(int tid = 0; tid < this->_nthreads; tid++) {
-			// Calculate bucket sizes by doing a binary search for each
-			// suffix and noting where it lands
-			try {
-				// Allocate and initialize containers for holding bucket
-				// sizes and representatives.
-				tparams[tid].bucketSzs.resize(numBuckets);
-				tparams[tid].bucketReps.resize(numBuckets);
-				for (size_t i = 0; i < tparams[tid].bucketSzs.size(); i++) {
-					tparams[tid].bucketSzs[i] = 0;
-					tparams[tid].bucketReps[i] = OFF_MASK;
-				}
-			} catch(bad_alloc &e) {
-				if(this->_passMemExc) {
-					throw e; // rethrow immediately
-				} else {
-					cerr << "Could not allocate sizes, representatives (" << ((numBuckets*8)>>10) << " KB) for blocks." << endl
-					     << "Please try using a smaller number of blocks by specifying a larger --bmax or a" << endl
-					     << "smaller --bmaxdivn." << endl;
-					throw 1;
-				}
-			}
-			tparams[tid].t = &t;
-			tparams[tid].sampleSuffs = &_sampleSuffs;
-			tparams[tid].begin = (tid == 0 ? 0 : len / this->_nthreads * tid);
-			tparams[tid].end = (tid + 1 == this->_nthreads ? len : len / this->_nthreads * (tid + 1));
-			if(this->_nthreads == 1) {
-				BinarySorting_worker<TStr>((void*)&tparams[tid]);
+		EList<TIndexOffU> bucketSzs(EBWTB_CAT); // holds computed bucket sizes
+		EList<TIndexOffU> bucketReps(EBWTB_CAT); // holds 1 member of each bucket (for splitting)
+		try {
+			// Allocate and initialize containers for holding bucket
+			// sizes and representatives.
+			bucketSzs.resizeExact(numBuckets);
+			bucketReps.resizeExact(numBuckets);
+			bucketSzs.fillZero();
+			bucketReps.fill(OFF_MASK);
+		} catch(bad_alloc &e) {
+			if(this->_passMemExc) {
+				throw e; // rethrow immediately
 			} else {
-#if (__cplusplus >= 201103l)
-				threads[tid] = new std::thread(BinarySorting_worker<TStr>, (void*)&tparams[tid]);
-#else
-				threads[tid] = new tthread::thread(BinarySorting_worker<TStr>, (void*)&tparams[tid]);
-#endif
+				cerr << "Could not allocate sizes, representatives (" << ((numBuckets*8)>>10) << " KB) for blocks." << endl
+				     << "Please try using a smaller number of blocks by specifying a larger --bmax or a" << endl
+				     << "smaller --bmaxdivn." << endl;
+				throw 1;
 			}
 		}
-		if(this->_nthreads > 1) {
-			for (int tid = 0; tid < this->_nthreads; tid++) {
-				threads[tid]->join();
-			}
-		}
-
-
-		EList<TIndexOffU>& bucketSzs = tparams[0].bucketSzs;
-		EList<TIndexOffU>& bucketReps = tparams[0].bucketReps;
-		for(int tid = 1; tid < this->_nthreads; tid++) {
-			for(size_t j = 0; j < numBuckets; j++) {
-				bucketSzs[j] += tparams[tid].bucketSzs[j];
-				if(bucketReps[j] == OFF_MASK) {
-					bucketReps[j] = tparams[tid].bucketReps[j];
+		// Iterate through every suffix in the text, determine which
+		// bucket it falls into by doing a binary search across the
+		// sorted list of samples, and increment a counter associated
+		// with that bucket.  Also, keep one representative for each
+		// bucket so that we can split it later.  We loop in ten
+		// stretches so that we can print out a helpful progress
+		// message.  (This step can take a long time.)
+		{
+			VMSG_NL("  Binary sorting into buckets");
+			Timer timer(cout, "  Binary sorting into buckets time: ", this->verbose());
+			TIndexOffU lenDiv10 = (TIndexOffU)((len + 9) / 10);
+			for(TIndexOffU iten = 0, ten = 0; iten < len; iten += lenDiv10, ten++) {
+				TIndexOffU itenNext = iten + lenDiv10;
+				if(ten > 0) VMSG_NL("  " << (ten * 10) << "%");
+				for(TIndexOffU i = iten; i < itenNext && i < len; i++) {
+					TIndexOffU r = binarySASearch(t, i, _sampleSuffs);
+					if(r == std::numeric_limits<TIndexOffU>::max()) continue; // r was one of the samples
+					assert_lt(r, numBuckets);
+					bucketSzs[r]++;
+					assert_lt(bucketSzs[r], len);
+					if(bucketReps[r] == OFF_MASK ||
+					   (_randomSrc.nextU32() & 100) == 0)
+					{
+						bucketReps[r] = i; // clobbers previous one, but that's OK
+					}
 				}
 			}
+			VMSG_NL("  100%");
 		}
-
 		// Check for large buckets and mergeable pairs of small buckets
 		// and split/merge as necessary
 		TIndexOff added = 0;
@@ -682,19 +523,19 @@ void KarkkainenBlockwiseSA<TStr>::buildSamples() {
 			VMSG_NL("Splitting and merging");
 			for(TIndexOffU i = 0; i < numBuckets; i++) {
 				TIndexOffU mergedSz = bsz + 1;
-				assert(bucketSzs[i] == 0 || bucketReps[i] != OFF_MASK);
-				if(i < (TIndexOffU)numBuckets-1) {
+				assert(bucketSzs[(size_t)i] == 0 || bucketReps[(size_t)i] != OFF_MASK);
+				if(i < numBuckets-1) {
 					mergedSz = bucketSzs[(size_t)i] + bucketSzs[(size_t)i+1] + 1;
 				}
 				// Merge?
 				if(mergedSz <= bsz) {
-					bucketSzs[i+1] += (bucketSzs[i]+1);
+					bucketSzs[(size_t)i+1] += (bucketSzs[(size_t)i]+1);
 					// The following may look strange, but it's necessary
 					// to ensure that the merged bucket has a representative
-					bucketReps[i+1] = _sampleSuffs[i+added];
-					erase(_sampleSuffs, i+added);
-					erase(bucketSzs, i);
-					erase(bucketReps, i);
+					bucketReps[(size_t)i+1] = _sampleSuffs[(size_t)i+added];
+					_sampleSuffs.erase((size_t)i+added);
+					bucketSzs.erase((size_t)i);
+					bucketReps.erase((size_t)i);
 					i--; // might go to -1 but ++ will overflow back to 0
 					numBuckets--;
 					merged++;
@@ -702,7 +543,7 @@ void KarkkainenBlockwiseSA<TStr>::buildSamples() {
 					assert_eq(numBuckets, bucketSzs.size());
 				}
 				// Split?
-				else if(bucketSzs[i] > bsz) {
+				else if(bucketSzs[(size_t)i] > bsz) {
 					// Add an additional sample from the bucketReps[]
 					// set accumulated in the binarySASearch loop; this
 					// effectively splits the bucket
@@ -724,7 +565,7 @@ void KarkkainenBlockwiseSA<TStr>::buildSamples() {
 		// split
 		VMSG_NL("Split " << added << ", merged " << merged << "; iterating...");
 	}
-// Do *not* force a do-over
+	// Do *not* force a do-over
 //	if(limit == 0) {
 //		VMSG_NL("Iterated too many times; trying again...");
 //		buildSamples();
@@ -753,17 +594,17 @@ static TIndexOffU suffixLcp(const T& t, TIndexOffU aOff, TIndexOffU bOff) {
  */
 template<typename TStr> inline
 bool KarkkainenBlockwiseSA<TStr>::tieBreakingLcp(TIndexOffU aOff,
-						 TIndexOffU bOff,
-						 TIndexOffU& lcp,
-						 bool& lcpIsSoft)
+                                                 TIndexOffU bOff,
+                                                 TIndexOffU& lcp,
+                                                 bool& lcpIsSoft)
 {
 	const TStr& t = this->text();
 	TIndexOffU c = 0;
-	TIndexOffU tlen = TIndexOffU(t.length());
+	TIndexOffU tlen = (TIndexOffU)t.length();
 	assert_leq(aOff, tlen);
 	assert_leq(bOff, tlen);
-	assert(_dc != NULL);
-	uint32_t dcDist = _dc->tieBreakOff(aOff, bOff);
+	assert(_dc.get() != NULL);
+	uint32_t dcDist = _dc.get()->tieBreakOff(aOff, bOff);
 	lcpIsSoft = false; // hard until proven soft
 	while(c < dcDist &&    // we haven't hit the tie breaker
 	      c < tlen-aOff && // we haven't fallen off of LHS suffix
@@ -781,7 +622,7 @@ bool KarkkainenBlockwiseSA<TStr>::tieBreakingLcp(TIndexOffU aOff,
 		// Hit a tie-breaker element
 		lcpIsSoft = true;
 		assert_neq(dcDist, 0xffffffff);
-		return _dc->breakTie(aOff+c, bOff+c) < 0;
+		return _dc.get()->breakTie(aOff+c, bOff+c) < 0;
 	} else {
 		assert_neq(t[aOff+c], t[bOff+c]);
 		return t[aOff+c] < t[bOff+c];
@@ -822,7 +663,7 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
 	const EList<TIndexOffU>& z)
 {
 	const TStr& t = this->text();
-	TIndexOffU len = TIndexOffU(t.length());
+	TIndexOffU len = (TIndexOffU)t.length();
 	// i is not covered by any previous match
 	TIndexOffU l;
 	if((int64_t)i > k) {
@@ -836,7 +677,7 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
 		assert_gt((int64_t)i, j);
 		TIndexOffU zIdx = (TIndexOffU)(i-j);
 		assert_leq(zIdx, len-cmp);
-		if(zIdx < _dcV || _dc == NULL) {
+		if(zIdx < _dcV || _dc.get() == NULL) {
 			// Go as far as the Z-box says
 			l = lookupSuffixZ(t, zIdx, cmp, z);
 			if(i + l > len) {
@@ -904,10 +745,10 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
 
 	// Now we're ready to do a comparison on the next char
 	if(l+i != len && (
-		   l == len-cmp || // departure from paper algorithm:
-		   // falling off pattern implies
-		   // pattern is *greater* in our case
-		   t[i + l] < t[cmp + l]))
+	   l == len-cmp || // departure from paper algorithm:
+	                   // falling off pattern implies
+	                   // pattern is *greater* in our case
+	   t[i + l] < t[cmp + l]))
 	{
 		// Case 2: Text suffix is less than upper sample suffix
 #ifndef NDEBUG
@@ -915,7 +756,6 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
 			assert(sstr_suf_lt(t, i, t, cmp, false));
 		}
 #endif
-
 		return true; // suffix at i is less than suffix at cmp
 	}
 	else {
@@ -925,7 +765,6 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
 			assert(sstr_suf_gt(t, i, t, cmp, false));
 		}
 #endif
-
 		return false; // suffix at i is less than suffix at cmp
 	}
 }
@@ -935,20 +774,12 @@ bool KarkkainenBlockwiseSA<TStr>::suffixCmp(
  * of the blockwise suffix sorting process.
  */
 template<typename TStr>
-void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
-#ifndef NDEBUG
-	if(this->_nthreads > 1) {
-		assert_lt((size_t)tid, this->_itrBuckets.size());
-	}
-#endif
-	EList<TIndexOffU>& bucket = (this->_nthreads > 1 ? this->_itrBuckets[tid] : this->_itrBucket);
-	{
-		ThreadSafe ts(&_mutex);
-		VMSG_NL("Getting block " << (cur_block+1) << " of " << _sampleSuffs.size()+1);
-	}
+void KarkkainenBlockwiseSA<TStr>::nextBlock() {
+	EList<TIndexOffU>& bucket = this->_itrBucket;
+	VMSG_NL("Getting block " << (_cur+1) << " of " << _sampleSuffs.size()+1);
 	assert(_built);
 	assert_gt(_dcV, 3);
-	assert_leq((size_t)cur_block, _sampleSuffs.size());
+	assert_leq(_cur, _sampleSuffs.size());
 	const TStr& t = this->text();
 	TIndexOffU len = (TIndexOffU)t.length();
 	// Set up the bucket
@@ -957,11 +788,8 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 	if(_sampleSuffs.size() == 0) {
 		// Special case: if _sampleSuffs is 0, then multikey-quicksort
 		// everything
-		{
-			ThreadSafe ts(&_mutex);
-			VMSG_NL("  No samples; assembling all-inclusive block");
-		}
-		assert_eq(0, cur_block);
+		VMSG_NL("  No samples; assembling all-inclusive block");
+		assert_eq(0, _cur);
 		try {
 			if(bucket.capacity() < this->bucketSz()) {
 				bucket.reserveExact(len+1);
@@ -982,10 +810,7 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 		}
 	} else {
 		try {
-			{
-				ThreadSafe ts(&_mutex);
-				VMSG_NL("  Reserving size (" << this->bucketSz() << ") for bucket " << (cur_block+1));
-			}
+			VMSG_NL("  Reserving size (" << this->bucketSz() << ") for bucket");
 			// BTL: Add a +100 fudge factor; there seem to be instances
 			// where a bucket ends up having one more elt than bucketSz()
 			if(bucket.size() < this->bucketSz()+100) {
@@ -1004,23 +829,18 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 		// Select upper and lower bounds from _sampleSuffs[] and
 		// calculate the Z array up to the difference-cover periodicity
 		// for both.  Be careful about first/last buckets.
-
-		//String<TIndexOffU> zLo(EBWTB_CAT), zHi(EBWTB_CAT);
-		EList<TIndexOffU> zLo, zHi;
-		assert_geq(cur_block, 0);
-		assert_leq((size_t)cur_block, _sampleSuffs.size());
-		bool first = (cur_block == 0);
-		bool last  = ((size_t)cur_block == _sampleSuffs.size());
+		EList<TIndexOffU> zLo(EBWTB_CAT), zHi(EBWTB_CAT);
+		assert_geq(_cur, 0);
+		assert_leq(_cur, _sampleSuffs.size());
+		bool first = (_cur == 0);
+		bool last  = (_cur == _sampleSuffs.size());
 		try {
-			// Timer timer(cout, "  Calculating Z arrays time: ", this->verbose());
-			{
-				ThreadSafe ts(&_mutex);
-				VMSG_NL("  Calculating Z arrays for bucket " << (cur_block+1));
-			}
+			Timer timer(cout, "  Calculating Z arrays time: ", this->verbose());
+			VMSG_NL("  Calculating Z arrays");
 			if(!last) {
 				// Not the last bucket
-				assert_lt((size_t)cur_block, _sampleSuffs.size());
-				hi = _sampleSuffs[cur_block];
+				assert_lt(_cur, _sampleSuffs.size());
+				hi = _sampleSuffs[_cur];
 				zHi.resizeExact(_dcV);
 				zHi.fillZero();
 				assert_eq(zHi[0], 0);
@@ -1028,9 +848,9 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 			}
 			if(!first) {
 				// Not the first bucket
-				assert_gt(cur_block, 0);
-				assert_leq((size_t)cur_block, _sampleSuffs.size());
-				lo = _sampleSuffs[cur_block-1];
+				assert_gt(_cur, 0);
+				assert_leq(_cur, _sampleSuffs.size());
+				lo = _sampleSuffs[_cur-1];
 				zLo.resizeExact(_dcV);
 				zLo.fillZero();
 				assert_gt(_dcV, 3);
@@ -1060,61 +880,49 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 		bool kHiSoft = false, kLoSoft = false;
 		assert_eq(0, bucket.size());
 		{
-			// Timer timer(cout, "  Block accumulator loop time: ", this->verbose());
-			{
-				ThreadSafe ts(&_mutex);
-				VMSG_NL("  Entering block accumulator loop for bucket " << (cur_block+1) << ":");
-			}
+			Timer timer(cout, "  Block accumulator loop time: ", this->verbose());
+			VMSG_NL("  Entering block accumulator loop:");
 			TIndexOffU lenDiv10 = (len + 9) / 10;
 			for(TIndexOffU iten = 0, ten = 0; iten < len; iten += lenDiv10, ten++) {
-				TIndexOffU itenNext = iten + lenDiv10;
-				{
-					ThreadSafe ts(&_mutex);
-					if(ten > 0) VMSG_NL("  bucket " << (cur_block+1) << ": " << (ten * 10) << "%");
+			TIndexOffU itenNext = iten + lenDiv10;
+			if(ten > 0) VMSG_NL("  " << (ten * 10) << "%");
+			for(TIndexOffU i = iten; i < itenNext && i < len; i++) {
+				assert_lt(jLo, (TIndexOff)i); assert_lt(jHi, (TIndexOff)i);
+				// Advance the upper-bound comparison by one character
+				if(i == hi || i == lo) continue; // equal to one of the bookends
+				if(hi != OFF_MASK && !suffixCmp(hi, i, jHi, kHi, kHiSoft, zHi)) {
+					continue; // not in the bucket
 				}
-				for(TIndexOffU i = iten; i < itenNext && i < len; i++) {
-					assert_lt(jLo, (TIndexOff)i); assert_lt(jHi, (TIndexOff)i);
-					// Advance the upper-bound comparison by one character
-					if(i == hi || i == lo) continue; // equal to one of the bookends
-					if(hi != OFF_MASK && !suffixCmp(hi, i, jHi, kHi, kHiSoft, zHi)) {
-						continue; // not in the bucket
-					}
-					if(lo != OFF_MASK && suffixCmp(lo, i, jLo, kLo, kLoSoft, zLo)) {
-						continue; // not in the bucket
-					}
-					// In the bucket! - add it
-					assert_lt(i, len);
-					try {
-						bucket.push_back(i);
-					} catch(bad_alloc &e) {
-						cerr << "Could not append element to block of " << ((bucket.size()) * OFF_SIZE) << " bytes" << endl;
-						if(this->_passMemExc) {
-							throw e; // rethrow immediately
-						} else {
-							cerr << "Please try using a larger number of blocks by specifying a smaller --bmax or" << endl
-							     << "a larger --bmaxdivn" << endl;
-							throw 1;
-						}
-					}
-					// Not necessarily true; we allow overflowing buckets
-					// since we can't guarantee that a good set of sample
-					// suffixes can be found in a reasonable amount of time
-					//assert_lt(bucket.size(), this->bucketSz());
+				if(lo != OFF_MASK && suffixCmp(lo, i, jLo, kLo, kLoSoft, zLo)) {
+					continue; // not in the bucket
 				}
-			} // end loop over all suffixes of t
-			{
-				ThreadSafe ts(&_mutex);
-				VMSG_NL("  bucket " << (cur_block+1) << ": 100%");
+				// In the bucket! - add it
+				assert_lt(i, len);
+				try {
+					bucket.push_back(i);
+				} catch(bad_alloc &e) {
+					cerr << "Could not append element to block of " << ((bucket.size()) * OFF_SIZE) << " bytes" << endl;
+					if(this->_passMemExc) {
+						throw e; // rethrow immediately
+					} else {
+						cerr << "Please try using a larger number of blocks by specifying a smaller --bmax or" << endl
+						     << "a larger --bmaxdivn" << endl;
+						throw 1;
+					}
+				}
+				// Not necessarily true; we allow overflowing buckets
+				// since we can't guarantee that a good set of sample
+				// suffixes can be found in a reasonable amount of time
+				//assert_lt(bucket.size(), this->bucketSz());
 			}
+			} // end loop over all suffixes of t
+			VMSG_NL("  100%");
 		}
 	} // end else clause of if(_sampleSuffs.size() == 0)
 	// Sort the bucket
 	if(bucket.size() > 0) {
 		Timer timer(cout, "  Sorting block time: ", this->verbose());
-		{
-			ThreadSafe ts(&_mutex);
-			VMSG_NL("  Sorting block of length " << bucket.size() << " for bucket " << (cur_block+1));
-		}
+		VMSG_NL("  Sorting block of length " << bucket.size());
 		this->qsort(bucket);
 	}
 	if(hi != OFF_MASK) {
@@ -1124,11 +932,8 @@ void KarkkainenBlockwiseSA<TStr>::nextBlock(int cur_block, int tid) {
 		// Final bucket; throw in $ suffix
 		bucket.push_back(len);
 	}
-	{
-		ThreadSafe ts(&_mutex);
-		VMSG_NL("Returning block of " << bucket.size() << " for bucket " << (cur_block+1));
-	}
-
+	VMSG_NL("Returning block of " << bucket.size());
+	_cur++; // advance to next bucket
 }
 
 #endif /*BLOCKWISE_SA_H_*/
