@@ -1,27 +1,43 @@
- #ifndef MULTIKEY_QSORT_H_
+/*
+ * Copyright 2011, Ben Langmead <langmea@cs.jhu.edu>
+ *
+ * This file is part of Bowtie 2.
+ *
+ * Bowtie 2 is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
+ *
+ * Bowtie 2 is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with Bowtie 2.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#ifndef MULTIKEY_QSORT_H_
 #define MULTIKEY_QSORT_H_
 
 #include <iostream>
-
+#include "sequence_io.h"
 #include "alphabet.h"
 #include "assert_helpers.h"
-#include "btypes.h"
 #include "diff_sample.h"
-#include "ds.h"
-#include "sequence_io.h"
+#include "sstring.h"
+#include "btypes.h"
 
 using namespace std;
 
 /**
- * Swap elements a and b in seqan::String s
+ * Swap elements a and b in s
  */
 template <typename TStr, typename TPos>
 static inline void swap(TStr& s, size_t slen, TPos a, TPos b) {
 	assert_lt(a, slen);
 	assert_lt(b, slen);
-	char tmp = s[a];
-	s[a] = s[b];
-	s[b] = tmp;
+	swap(s[a], s[b]);
 }
 
 /**
@@ -31,15 +47,13 @@ template <typename TVal, typename TPos>
 static inline void swap(TVal* s, size_t slen, TPos a, TPos b) {
 	assert_lt(a, slen);
 	assert_lt(b, slen);
-	TVal tmp = s[a];
-	s[a] = s[b];
-	s[b] = tmp;
+	swap(s[a], s[b]);
 }
 
 /**
- * Helper macro for swapping elements a and b in seqan::String s.  Does
- * some additional sainty checking w/r/t begin and end (which are
- * parameters to the sorting routines below).
+ * Helper macro for swapping elements a and b in s.  Does some additional
+ * sainty checking w/r/t begin and end (which are parameters to the sorting
+ * routines below).
  */
 #define SWAP(s, a, b) { \
 	assert_geq(a, begin); \
@@ -50,11 +64,11 @@ static inline void swap(TVal* s, size_t slen, TPos a, TPos b) {
 }
 
 /**
- * Helper macro for swapping the same pair of elements a and b in
- * two different seqan::Strings s and s2.  This is a helpful variant
- * if, for example, the caller would like to see how their input was
- * permuted by the sort routine (in that case, the caller would let s2
- * be an array s2[] where s2 is the same length as s and s2[i] = i).
+ * Helper macro for swapping the same pair of elements a and b in two different
+ * strings s and s2.  This is a helpful variant if, for example, the caller
+ * would like to see how their input was permuted by the sort routine (in that
+ * case, the caller would let s2 be an array s2[] where s2 is the same length
+ * as s and s2[i] = i).
  */
 #define SWAP2(s, s2, a, b) { \
 	SWAP(s, a, b); \
@@ -67,7 +81,7 @@ static inline void swap(TVal* s, size_t slen, TPos a, TPos b) {
 
 /**
  * Helper macro that swaps a range of elements [i, i+n) with another
- * range [j, j+n) in seqan::String s.
+ * range [j, j+n) in s.
  */
 #define VECSWAP(s, i, j, n) { \
 	if(n > 0) { vecswap(s, slen, i, j, n, begin, end); } \
@@ -75,7 +89,7 @@ static inline void swap(TVal* s, size_t slen, TPos a, TPos b) {
 
 /**
  * Helper macro that swaps a range of elements [i, i+n) with another
- * range [j, j+n) both in seqan::String s and seqan::String s2.
+ * range [j, j+n) both in s and s2.
  */
 #define VECSWAP2(s, s2, i, j, n) { \
 	if(n > 0) { vecswap2(s, slen, s2, i, j, n, begin, end); } \
@@ -83,9 +97,9 @@ static inline void swap(TVal* s, size_t slen, TPos a, TPos b) {
 
 /**
  * Helper function that swaps a range of elements [i, i+n) with another
- * range [j, j+n) in seqan::String s.  begin and end represent the
- * current range under consideration by the caller (one of the
- * recursive multikey_quicksort routines below).
+ * range [j, j+n) in s.  begin and end represent the current range under
+ * consideration by the caller (one of the recursive multikey_quicksort
+ * routines below).
  */
 template <typename TStr, typename TPos>
 static inline void vecswap(TStr& s, size_t slen, TPos i, TPos j, TPos n, TPos begin, TPos end) {
@@ -124,13 +138,22 @@ static inline void vecswap(TVal *s, size_t slen, TPos i, TPos j, TPos n, TPos be
 }
 
 /**
- * Helper function that swaps a range of elements [i, i+n) with another
- * range [j, j+n) both in seqan::String s and seqan::String s2.  begin
- * and end represent the current range under consideration by the
- * caller (one of the recursive multikey_quicksort routines below).
+ * Helper function that swaps a range of elements [i, i+n) with another range
+ * [j, j+n) both in s and s2.  begin and end represent the current range under
+ * consideration by the caller (one of the recursive multikey_quicksort
+ * routines below).
  */
 template <typename TStr, typename TPos>
-static inline void vecswap2(TStr& s, size_t slen, TStr& s2, TPos i, TPos j, TPos n, TPos begin, TPos end) {
+static inline void vecswap2(
+	TStr& s,
+	size_t slen,
+	TStr& s2,
+	TPos i,
+	TPos j,
+	TPos n,
+	TPos begin,
+	TPos end)
+{
 	assert_geq(i, begin);
 	assert_geq(j, begin);
 	assert_lt(i, end);
@@ -170,12 +193,13 @@ static inline void vecswap2(TVal* s, size_t slen, TVal* s2, TPos i, TPos j, TPos
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
 /// than any TAlphabet character - 'hi'.
-#define CHAR_AT(ss, aa) ((s[ss].length() > aa) ? (int)(s[ss][aa]) : hi)
+#define CHAR_AT(ss, aa) ((length(s[ss]) > aa) ? (int)(s[ss][aa]) : hi)
 
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
 /// than any TAlphabet character - 'hi'.
-#define CHAR_AT_SUF(si, off) (((off+s[si]) < hlen) ? ((int)(host[off+s[si]])) : (hi))
+#define CHAR_AT_SUF(si, off) \
+	(((off + s[si]) < hlen) ? ((int)(host[off + s[si]])) : (hi))
 
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
@@ -292,9 +316,9 @@ bool assertPartitionedSuf2(
 #endif
 
 /**
- * Assert that the seqan::String s of suffix offsets into seqan::String
- * 'host' is a seemingly legitimate suffix-offset list (at this time,
- * we just check that it doesn't list any suffix twice).
+ * Assert that string s of suffix offsets into string 'host' is a seemingly
+ * legitimate suffix-offset list (at this time, we just check that it doesn't
+ * list any suffix twice).
  */
 static inline void sanityCheckInputSufs(TIndexOffU *s, size_t slen) {
 	assert_gt(slen, 0);
@@ -310,17 +334,18 @@ static inline void sanityCheckInputSufs(TIndexOffU *s, size_t slen) {
 }
 
 /**
- * Assert that the seqan::String s of suffix offsets into seqan::String
- * 'host' really are in lexicographical order up to depth 'upto'.
+ * Assert that the string s of suffix offsets into  'host' really are in
+ * lexicographical order up to depth 'upto'.
  */
 template <typename T>
-void sanityCheckOrderedSufs(const T& host,
-                            size_t hlen,
-                            TIndexOffU *s,
-                            size_t slen,
-                            size_t upto,
-                            size_t lower = 0,
-                            size_t upper = OFF_MASK)
+void sanityCheckOrderedSufs(
+	const T& host,
+	size_t hlen,
+	TIndexOffU *s,
+	size_t slen,
+	size_t upto,
+	size_t lower = 0,
+	size_t upper = OFF_MASK)
 {
 	assert_lt(s[0], hlen);
 	upper = min<size_t>(upper, slen-1);
@@ -471,12 +496,13 @@ void mkeyQSortSuf(
 	if(sanityCheck) sanityCheckOrderedSufs(host, hlen, s, slen, upto);
 }
 
-struct QSortRange {
-	size_t begin;
-	size_t end;
-	size_t depth;
-};
-
+/**
+ * Just like mkeyQSortSuf but all swaps are applied to s2 as well as s.
+ * This is a helpful variant if, for example, the caller would like to
+ * see how their input was permuted by the sort routine (in that case,
+ * the caller would let s2 be an array s2[] where s2 is the same length
+ * as s and s2[i] = i).
+ */
 template<typename T>
 void mkeyQSortSuf2(
 	const T& host,
@@ -485,69 +511,45 @@ void mkeyQSortSuf2(
 	size_t slen,
 	TIndexOffU *s2,
 	int hi,
-	size_t _begin,
-	size_t _end,
-	size_t _depth,
-	size_t upto = OFF_MASK,
-	EList<size_t>* boundaries = NULL)
+	size_t begin,
+	size_t end,
+	size_t depth,
+	size_t upto = OFF_MASK)
 {
-	EList<EList<QSortRange> > block_list;
+	// Helper for making the recursive call; sanity-checks arguments to
+	// make sure that the problem actually got smaller.
+	#define MQS_RECURSE_SUF_DS(nbegin, nend, ndepth) { \
+		assert(nbegin > begin || nend < end || ndepth > depth); \
+		if(ndepth < upto) { /* don't exceed depth of 'upto' */ \
+			mkeyQSortSuf2(host, hlen, s, slen, s2, hi, nbegin, nend, ndepth, upto); \
+		} \
+	}
+	assert_leq(begin, slen);
+	assert_leq(end, slen);
+	size_t a, b, c, d, /*e,*/ r;
+	size_t n = end - begin;
+	if(n <= 1) return;                 // 1-element list already sorted
+	CHOOSE_AND_SWAP_PIVOT(SWAP2, CHAR_AT_SUF); // pick pivot, swap it into [begin]
+	int v = CHAR_AT_SUF(begin, depth); // v <- randomly-selected pivot value
+	#ifndef NDEBUG
+	{
+		bool stillInBounds = false;
+		for(size_t i = begin; i < end; i++) {
+			if(depth < (hlen-s[i])) {
+				stillInBounds = true;
+				break;
+			} else { /* already fell off this suffix */ }
+		}
+		assert(stillInBounds); // >=1 suffix must still be in bounds
+	}
+	#endif
+	a = b = begin;
+	c = d = /*e =*/ end-1;
 	while(true) {
-		size_t begin = 0, end = 0, depth = 0;
-		if(block_list.size() == 0) {
-			begin = _begin;
-			end = _end;
-			depth = _depth;
-		} else {
-			if(block_list.back().size() > 0) {
-				begin = block_list.back()[0].begin;
-				end = block_list.back()[0].end;
-				depth = block_list.back()[0].depth;
-				block_list.back().erase(0);
-			} else {
-				block_list.resize(block_list.size() - 1);
-				if(block_list.size() == 0) {
-					break;
-				}
-			}
-		}
-		if(depth == upto) {
-			if(boundaries != NULL) {
-				(*boundaries).push_back(end);
-			}
-			continue;
-		}
-		assert_leq(begin, slen);
-		assert_leq(end, slen);
-		size_t a, b, c, d, /*e,*/ r;
-		size_t n = end - begin;
-		if(n <= 1) { // 1-element list already sorted
-			if(n == 1 && boundaries != NULL) {
-				boundaries->push_back(end);
-			}
-			continue;
-		}
-		CHOOSE_AND_SWAP_PIVOT(SWAP2, CHAR_AT_SUF); // pick pivot, swap it into [begin]
-		int v = CHAR_AT_SUF(begin, depth); // v <- randomly-selected pivot value
-#ifndef NDEBUG
-		{
-			bool stillInBounds = false;
-			for(size_t i = begin; i < end; i++) {
-				if(depth < (hlen-s[i])) {
-					stillInBounds = true;
-					break;
-				} else { /* already fell off this suffix */ }
-			}
-			assert(stillInBounds); // >=1 suffix must still be in bounds
-		}
-#endif
-		a = b = begin;
-		c = d = /*e =*/ end-1;
-		while(true) {
-			// Invariant: everything before a is = pivot, everything
-			// between a and b is <
-			int bc = 0; // shouldn't have to init but gcc on Mac complains
-			while(b <= c && v >= (bc = CHAR_AT_SUF(b, depth))) {
+		// Invariant: everything before a is = pivot, everything
+		// between a and b is <
+		int bc = 0; // shouldn't have to init but gcc on Mac complains
+		while(b <= c && v >= (bc = CHAR_AT_SUF(b, depth))) {
 			if(v == bc) {
 				SWAP2(s, s2, a, b); a++;
 			}
@@ -564,49 +566,32 @@ void mkeyQSortSuf2(
 			c--;
 		}
 		if(b > c) break;
-			SWAP2(s, s2, b, c);
-			b++;
-			c--;
-		}
-		assert(a > begin || c < end-1);                      // there was at least one =s
-		assert_lt(/*e*/d-c, n); // they can't all have been > pivot
-		assert_lt(b-a, n); // they can't all have been < pivot
-		assert(assertPartitionedSuf(host, s, slen, hi, v, begin, end, depth));  // check pre-=-swap invariant
-		r = min(a-begin, b-a); VECSWAP2(s, s2, begin, b-r,   r);  // swap left = to center
-		r = min(d-c, end-d-1); VECSWAP2(s, s2, b,     end-r, r);  // swap right = to center
-		assert(assertPartitionedSuf2(host, s, slen, hi, v, begin, end, depth)); // check post-=-swap invariant
-		r = b-a; // r <- # of <'s
-		EList<QSortRange> tmp1;
-		block_list.push_back(tmp1);
-		block_list.back().clear();
-		if(r > 0) { // recurse on <'s
-			QSortRange tmp2;
-			block_list.back().push_back(tmp2);
-			block_list.back().back().begin = begin;
-			block_list.back().back().end = begin + r;
-			block_list.back().back().depth = depth;
-		}
-		// Do not recurse on ='s if the pivot was the off-the-end value;
-		// they're already fully sorted
-		//if(v != hi) { // recurse on ='s
-			QSortRange tmp3;
-			block_list.back().push_back(tmp3);
-			block_list.back().back().begin = begin + r;
-			block_list.back().back().end = begin + r + (a-begin) + (end-d-1);
-			block_list.back().back().depth = depth + 1;
-			//}
-		r = d-c;   // r <- # of >'s excluding those exhausted
-		if(r > 0 /*&& v < hi-1*/) { // recurse on >'s
-			QSortRange tmp4;
-			block_list.back().push_back(tmp4);
-			block_list.back().back().begin = end - r;
-			block_list.back().back().end = end;
-			block_list.back().back().depth = depth;
-		}
+		SWAP2(s, s2, b, c);
+		b++;
+		c--;
+	}
+	assert(a > begin || c < end-1);                      // there was at least one =s
+	assert_lt(/*e*/d-c, n); // they can't all have been > pivot
+	assert_lt(b-a, n); // they can't all have been < pivot
+	assert(assertPartitionedSuf(host, s, slen, hi, v, begin, end, depth));  // check pre-=-swap invariant
+	r = min(a-begin, b-a); VECSWAP2(s, s2, begin, b-r,   r);  // swap left = to center
+	r = min(d-c, end-d-1); VECSWAP2(s, s2, b,     end-r, r);  // swap right = to center
+	assert(assertPartitionedSuf2(host, s, slen, hi, v, begin, end, depth)); // check post-=-swap invariant
+	r = b-a; // r <- # of <'s
+	if(r > 0) {
+		MQS_RECURSE_SUF_DS(begin, begin + r, depth); // recurse on <'s
+	}
+	// Do not recurse on ='s if the pivot was the off-the-end value;
+	// they're already fully sorted
+	if(v != hi) {
+		MQS_RECURSE_SUF_DS(begin + r, begin + r + (a-begin) + (end-d-1), depth+1); // recurse on ='s
+	}
+	r = d-c;   // r <- # of >'s excluding those exhausted
+	if(r > 0 && v < hi-1) {
+		MQS_RECURSE_SUF_DS(end-r, end, depth); // recurse on >'s
 	}
 }
 
-
 /**
  * Toplevel function for multikey quicksort over suffixes with double
  * swapping.
@@ -620,8 +605,7 @@ void mkeyQSortSuf2(
 	int hi,
 	bool verbose = false,
 	bool sanityCheck = false,
-	size_t upto = OFF_MASK,
-	EList<size_t>* boundaries = NULL)
+	size_t upto = OFF_MASK)
 {
 	size_t hlen = host.length();
 	if(sanityCheck) sanityCheckInputSufs(s, slen);
@@ -630,7 +614,7 @@ void mkeyQSortSuf2(
 		sOrig = new TIndexOffU[slen];
 		memcpy(sOrig, s, OFF_SIZE * slen);
 	}
-	mkeyQSortSuf2(host, hlen, s, slen, s2, hi, (size_t)0, slen, (size_t)0, upto, boundaries);
+	mkeyQSortSuf2(host, hlen, s, slen, s2, hi, (size_t)0, slen, (size_t)0, upto);
 	if(sanityCheck) {
 		sanityCheckOrderedSufs(host, hlen, s, slen, upto);
 		for(size_t i = 0; i < slen; i++) {
@@ -649,16 +633,15 @@ class DifferenceCoverSample;
  * Constant time
  */
 template<typename T1, typename T2> inline
-bool sufDcLt(const T1& host,
-             const T2& s1,
-             const T2& s2,
-             const DifferenceCoverSample<T1>& dc,
-             bool sanityCheck = false)
+bool sufDcLt(
+	const T1& host,
+	const T2& s1,
+	const T2& s2,
+	const DifferenceCoverSample<T1>& dc,
+	bool sanityCheck = false)
 {
 	size_t diff = dc.tieBreakOff(s1, s2);
-#ifndef NDEBUG
-	size_t hlen = host.length();
-#endif
+	ASSERT_ONLY(size_t hlen = host.length());
 	assert_lt(diff, dc.v());
 	assert_lt(diff, hlen-s1);
 	assert_lt(diff, hlen-s2);
@@ -672,7 +655,6 @@ bool sufDcLt(const T1& host,
 	if(sanityCheck && ret != sstr_suf_lt(host, s1, hlen, host, s2, hlen, false)) {
 		assert(false);
 	}
-
 #endif
 	return ret;
 }
@@ -681,14 +663,15 @@ bool sufDcLt(const T1& host,
  * k log(k)
  */
 template<typename T> inline
-void qsortSufDc(const T& host,
-                size_t hlen,
-                TIndexOffU* s,
-                size_t slen,
-                const DifferenceCoverSample<T>& dc,
-                size_t begin,
-                size_t end,
-                bool sanityCheck = false)
+void qsortSufDc(
+	const T& host,
+	size_t hlen,
+	TIndexOffU* s,
+	size_t slen,
+	const DifferenceCoverSample<T>& dc,
+	size_t begin,
+	size_t end,
+	bool sanityCheck = false)
 {
 	assert_leq(end, slen);
 	assert_lt(begin, slen);
@@ -706,8 +689,7 @@ void qsortSufDc(const T& host,
 	for(size_t i = begin; i < end-1; i++) {
 		if(sufDcLt(host, s[i], s[end-1], dc, sanityCheck)) {
 			if(sanityCheck)
-				assert(sstr_suf_lt(host, s[i], hlen, host, s[end-1], hlen, false));
-
+				assert(dollarLt(suffix(host, s[i]), suffix(host, s[end-1])));
 			assert_lt(begin + cur, end-1);
 			SWAP(s, i, begin + cur);
 			cur++;
@@ -724,15 +706,16 @@ void qsortSufDc(const T& host,
  * Toplevel function for multikey quicksort over suffixes.
  */
 template<typename T1, typename T2>
-void mkeyQSortSufDcU8(const T1& host1,
-                      const T2& host,
-                      size_t hlen,
-                      TIndexOffU* s,
-                      size_t slen,
-                      const DifferenceCoverSample<T1>& dc,
-                      int hi,
-                      bool verbose = false,
-                      bool sanityCheck = false)
+void mkeyQSortSufDcU8(
+	const T1& host1,
+	const T2& host,
+	size_t hlen,
+	TIndexOffU* s,
+	size_t slen,
+	const DifferenceCoverSample<T1>& dc,
+	int hi,
+	bool verbose = false,
+	bool sanityCheck = false)
 {
 	if(sanityCheck) sanityCheckInputSufs(s, slen);
 	mkeyQSortSufDcU8(host1, host, hlen, s, slen, dc, hi, 0, slen, 0, sanityCheck);
@@ -744,13 +727,14 @@ void mkeyQSortSufDcU8(const T1& host1,
  * cover to break the tie.
  */
 template<typename T1, typename T2> inline
-bool sufDcLtU8(const T1& host1,
-               const T2& host,
-               size_t hlen,
-               size_t s1,
-               size_t s2,
-               const DifferenceCoverSample<T1>& dc,
-               bool sanityCheck = false)
+bool sufDcLtU8(
+	const T1& host1,
+	const T2& host,
+	size_t hlen,
+	size_t s1,
+	size_t s2,
+	const DifferenceCoverSample<T1>& dc,
+	bool sanityCheck = false)
 {
 	hlen += 0;
 	size_t diff = dc.tieBreakOff((TIndexOffU)s1, (TIndexOffU)s2);
@@ -767,7 +751,6 @@ bool sufDcLtU8(const T1& host1,
 #ifndef NDEBUG
 	bool ret2 = sstr_suf_lt(host1, s1, hlen, host, s2, hlen, false);
 	assert(!sanityCheck || ret == ret2);
-
 #endif
 	return ret;
 }
@@ -776,15 +759,16 @@ bool sufDcLtU8(const T1& host1,
  * k log(k)
  */
 template<typename T1, typename T2> inline
-void qsortSufDcU8(const T1& host1,
-                  const T2& host,
-                  size_t hlen,
-                  TIndexOffU* s,
-                  size_t slen,
-                  const DifferenceCoverSample<T1>& dc,
-                  size_t begin,
-                  size_t end,
-                  bool sanityCheck = false)
+void qsortSufDcU8(
+	const T1& host1,
+	const T2& host,
+	size_t hlen,
+	TIndexOffU* s,
+	size_t slen,
+	const DifferenceCoverSample<T1>& dc,
+	size_t begin,
+	size_t end,
+	bool sanityCheck = false)
 {
 	assert_leq(end, slen);
 	assert_lt(begin, slen);
@@ -806,7 +790,6 @@ void qsortSufDcU8(const T1& host1,
 				assert(sstr_suf_lt(host1, s[i], hlen, host1, s[end-1], hlen, false));
 			}
 			assert_lt(begin + cur, end-1);
-
 #endif
 			SWAP(s, i, begin + cur);
 			cur++;
@@ -822,6 +805,9 @@ void qsortSufDcU8(const T1& host1,
 #define BUCKET_SORT_CUTOFF (4 * 1024 * 1024)
 #define SELECTION_SORT_CUTOFF 6
 
+// 5 64-element buckets for bucket-sorting A, C, G, T, $
+extern TIndexOffU bkts[4][4 * 1024 * 1024];
+
 /**
  * Straightforwardly obtain a uint8_t-ized version of t[off].  This
  * works fine as long as TStr is not packed.
@@ -846,12 +832,13 @@ inline uint8_t get_uint8<S2bDnaString>(const S2bDnaString& t, size_t off) {
  * 's' of suffixes.  If the character is out-of-bounds, return hi.
  */
 template<typename TStr>
-static inline int char_at_suf_u8(const TStr& host,
-                                 size_t hlen,
-                                 TIndexOffU* s,
-                                 size_t si,
-                                 size_t off,
-                                 uint8_t hi)
+static inline int char_at_suf_u8(
+	const TStr& host,
+	size_t hlen,
+	TIndexOffU* s,
+	size_t si,
+	size_t off,
+	uint8_t hi)
 {
 	return ((off+s[si]) < hlen) ? get_uint8(host, off+s[si]) : (hi);
 }
@@ -871,7 +858,7 @@ static void selectionSortSufDcU8(
         bool sanityCheck = false)
 {
 #define ASSERT_SUF_LT(l, r) \
-	if(sanityCheck &&						\
+	if(sanityCheck && \
 	   !sstr_suf_lt(host1, s[l], hlen, host1, s[r], hlen, false)) { \
 		assert(false); \
 	}
@@ -991,97 +978,70 @@ static void bucketSortSufDcU8(
         size_t slen,
         const DifferenceCoverSample<T1>& dc,
         uint8_t hi,
-        size_t _begin,
-        size_t _end,
-        size_t _depth,
+        size_t begin,
+        size_t end,
+        size_t depth,
         bool sanityCheck = false)
 {
-        // 5 64-element buckets for bucket-sorting A, C, G, T, $
-        TIndexOffU* bkts[4];
-	for(size_t i = 0; i < 4; i++) {
-	        bkts[i] = new TIndexOffU[4 * 1024 * 1024];
+	size_t cnts[] = { 0, 0, 0, 0, 0 };
+	#define BKT_RECURSE_SUF_DC_U8(nbegin, nend) { \
+		bucketSortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc, hi, \
+		                         (nbegin), (nend), depth+1, sanityCheck); \
 	}
-	EList<EList<size_t> > block_list;
-	bool first = true;
-	while(true) {
-	        size_t begin = 0, end = 0;
-		if(first) {
-		        begin = _begin;
-			end = _end;
-			first = false;
+	assert_gt(end, begin);
+	assert_leq(end-begin, BUCKET_SORT_CUTOFF);
+	assert_eq(hi, 4);
+	if(end == begin+1) return; // 1-element list already sorted
+	if(depth > dc.v()) {
+		// Quicksort the remaining suffixes using difference cover
+		// for constant-time comparisons; this is O(k*log(k)) where
+		// k=(end-begin)
+		qsortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc, begin, end, sanityCheck);
+		return;
+	}
+	if(end-begin <= SELECTION_SORT_CUTOFF) {
+		// Bucket sort remaining items
+		selectionSortSufDcU8(host1, host, hlen, s, slen, dc, hi,
+		                     begin, end, depth, sanityCheck);
+		if(sanityCheck) {
+			sanityCheckOrderedSufs(host1, hlen, s, slen,
+			                       OFF_MASK, begin, end);
+		}
+		return;
+	}
+	for(size_t i = begin; i < end; i++) {
+		size_t off = depth + s[i];
+		uint8_t c = (off < hlen) ? get_uint8(host, off) : hi;
+		assert_leq(c, 4);
+		if(c == 0) {
+			s[begin + cnts[0]++] = s[i];
 		} else {
-		        if(block_list.size() == 0) {
-			        break;
-			}
-			if(block_list.back().size() > 1) {
-			        end = block_list.back().back(); 
-				block_list.back().pop_back();
-				begin = block_list.back().back();
-			} else {
-			        block_list.resize(block_list.size() - 1);
-				if(block_list.size() == 0) {
-				        break;
-				}
-			}
-		}
-		size_t depth = block_list.size() + _depth;
-		assert_leq(end-begin, BUCKET_SORT_CUTOFF);
-		assert_eq(hi, 4);
-		if(end <= begin + 1) { // 1-element list already sorted
-		        continue;
-		}
-		if(depth > dc.v()) {
-		        // Quicksort the remaining suffixes using difference cover
-		        // for constant-time comparisons; this is O(k*log(k)) where
-		        // k=(end-begin)
-		        qsortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc, begin, end, sanityCheck);
-			continue;
-		}
-		if(end-begin <= SELECTION_SORT_CUTOFF) {
-		        // Bucket sort remaining items
-		        selectionSortSufDcU8(host1, host, hlen, s, slen, dc, hi,
-					     begin, end, depth, sanityCheck);
-			if(sanityCheck) {
-			        sanityCheckOrderedSufs(host1, hlen, s, slen,
-						       OFF_MASK, begin, end);
-			}
-			continue;
-		}
-		size_t cnts[] = { 0, 0, 0, 0, 0 };
-		for(size_t i = begin; i < end; i++) {
-		        size_t off = depth + s[i];
-			uint8_t c = (off < hlen) ? get_uint8(host, off) : hi;
-			assert_leq(c, 4);
-			if(c == 0) {
-			        s[begin + cnts[0]++] = s[i];
-			} else {
-			        bkts[c-1][cnts[c]++] = s[i];
-			}
-		}
-		assert_eq(cnts[0] + cnts[1] + cnts[2] + cnts[3] + cnts[4], end - begin);
-		size_t cur = begin + cnts[0];
-		if(cnts[1] > 0) { memcpy(&s[cur], bkts[0], cnts[1] << (OFF_SIZE/4 + 1)); cur += cnts[1]; }
-		if(cnts[2] > 0) { memcpy(&s[cur], bkts[1], cnts[2] << (OFF_SIZE/4 + 1)); cur += cnts[2]; }
-		if(cnts[3] > 0) { memcpy(&s[cur], bkts[2], cnts[3] << (OFF_SIZE/4 + 1)); cur += cnts[3]; }
-		if(cnts[4] > 0) { memcpy(&s[cur], bkts[3], cnts[4] << (OFF_SIZE/4 + 1)); }
-		// This frame is now totally finished with bkts[][], so recursive
-		// callees can safely clobber it; we're not done with cnts[], but
-		// that's local to the stack frame.
-		EList<size_t> tmp;
-		block_list.push_back(tmp);
-		block_list.back().clear();
-		block_list.back().push_back(begin);
-		for(size_t i = 0; i < 4; i++) {
-		        if(cnts[i] > 0) {
-			        block_list.back().push_back(block_list.back().back() + cnts[i]);
-			}
+			bkts[c-1][cnts[c]++] = s[i];
 		}
 	}
+	assert_eq(cnts[0] + cnts[1] + cnts[2] + cnts[3] + cnts[4], end - begin);
+	size_t cur = begin + cnts[0];
+	if(cnts[1] > 0) { memcpy(&s[cur], bkts[0], cnts[1] << (OFF_SIZE/4 + 1)); cur += cnts[1]; }
+	if(cnts[2] > 0) { memcpy(&s[cur], bkts[1], cnts[2] << (OFF_SIZE/4 + 1)); cur += cnts[2]; }
+	if(cnts[3] > 0) { memcpy(&s[cur], bkts[2], cnts[3] << (OFF_SIZE/4 + 1)); cur += cnts[3]; }
+	if(cnts[4] > 0) { memcpy(&s[cur], bkts[3], cnts[4] << (OFF_SIZE/4 + 1)); }
+	// This frame is now totally finished with bkts[][], so recursive
+	// callees can safely clobber it; we're not done with cnts[], but
+	// that's local to the stack frame.
+	cur = begin;
+	if(cnts[0] > 0) {
+		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[0]); cur += cnts[0];
+	}
+	if(cnts[1] > 0) {
+		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[1]); cur += cnts[1];
+	}
+	if(cnts[2] > 0) {
+		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[2]); cur += cnts[2];
+	}
+	if(cnts[3] > 0) {
+		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[3]);
+	}
 	// Done
-	
-	for(size_t i = 0; i < 4; i++) {
-	        delete [] bkts[i];
-	}
 }
 
 /**
@@ -1104,17 +1064,18 @@ static void bucketSortSufDcU8(
  *     bounds the depth to which the function sorts.
  */
 template<typename T1, typename T2>
-void mkeyQSortSufDcU8(const T1& host1,
-                      const T2& host,
-                      size_t hlen,
-                      TIndexOffU* s,
-                      size_t slen,
-                      const DifferenceCoverSample<T1>& dc,
-                      int hi,
-                      size_t begin,
-                      size_t end,
-                      size_t depth,
-                      bool sanityCheck = false)
+void mkeyQSortSufDcU8(
+	const T1& host1,
+	const T2& host,
+	size_t hlen,
+	TIndexOffU* s,
+	size_t slen,
+	const DifferenceCoverSample<T1>& dc,
+	int hi,
+	size_t begin,
+	size_t end,
+	size_t depth,
+	bool sanityCheck = false)
 {
 	// Helper for making the recursive call; sanity-checks arguments to
 	// make sure that the problem actually got smaller.
