@@ -1,4 +1,22 @@
-#include "alphabet.h"
+/*
+ * Copyright 2011, Ben Langmead <langmea@cs.jhu.edu>
+ *
+ * This file is part of Bowtie 2.
+ *
+ * Bowtie 2 is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
+ *
+ * Bowtie 2 is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with Bowtie 2.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
 #include "ref_read.h"
 
 /**
@@ -7,10 +25,11 @@
  * anything with the sequence characters themselves; this is purely for
  * measuring lengths.
  */
-RefRecord fastaRefReadSize(FileBuf& in,
-                           const RefReadInParams& rparms,
-                           bool first,
-                           BitpairOutFileBuf* bpout)
+RefRecord fastaRefReadSize(
+	FileBuf& in,
+	const RefReadInParams& rparms,
+	bool first,
+	BitpairOutFileBuf* bpout)
 {
 	int c;
 	static int lastc = '>'; // last character seen
@@ -65,27 +84,41 @@ RefRecord fastaRefReadSize(FileBuf& in,
 
 	// Now skip to the first DNA character, counting gap characters
 	// as we go
+	int lc = -1; // last-DNA char variable for color conversion
 	while(true) {
-		int cat = dna4Cat[c];
-		if(rparms.nsToAs && cat == 2) c = 'A';
+		int cat = asc2dnacat[c];
+		if(rparms.nsToAs && cat >= 2) c = 'A';
 		if(cat == 1) {
 			// This is a DNA character
-			break; // to read-in loop
-                } else if (cat == 2) {
+			if(rparms.color) {
+				if(lc != -1) {
+					// Got two consecutive unambiguous DNAs
+					break; // to read-in loop
+				}
+				// Keep going; we need two consecutive unambiguous DNAs
+				lc = asc2dna[(int)c];
+				// The 'if(off > 0)' takes care of the case where
+				// the reference is entirely unambiguous and we don't
+				// want to incorrectly increment off.
+				if(off > 0) off++;
+			} else {
+				break; // to read-in loop
+			}
+		} else if(cat >= 2) {
+			if(lc != -1 && off == 0) off++;
+			lc = -1;
 			off++; // skip over gap character and increment
-                } else if (c == '>') {
-                  if (off > 0 && lastc == '>') {
-                    cerr << "Warning: Encountered reference sequence with only "
-                            "gaps"
-                         << endl;
-                  } else if (lastc == '>') {
-                    cerr << "Warning: Encountered empty reference sequence"
-                         << endl;
-                  }
-                  lastc = '>';
-                  return RefRecord((TIndexOffU)off, 0, first);
-                }
-                c = in.get();
+		} else if(c == '>') {
+			if(off > 0 && lastc == '>') {
+				cerr << "Warning: Encountered reference sequence with only gaps" << endl;
+			} else if(lastc == '>') {
+				cerr << "Warning: Encountered empty reference sequence" << endl;
+			}
+			lastc = '>';
+			//return RefRecord(off, 0, false);
+			return RefRecord((TIndexOffU)off, 0, first);
+		}
+		c = in.get();
 		if(c == -1) {
 			// End-of-file
 			if(off > 0 && lastc == '>') {
@@ -94,29 +127,47 @@ RefRecord fastaRefReadSize(FileBuf& in,
 				cerr << "Warning: Encountered empty reference sequence" << endl;
 			}
 			lastc = -1;
+			//return RefRecord(off, 0, false);
 			return RefRecord((TIndexOffU)off, 0, first);
 		}
 	}
-	assert_eq(1, dna4Cat[c]); // C must be unambiguous base
+	assert(!rparms.color || (lc != -1));
+	assert_eq(1, asc2dnacat[c]); // C must be unambiguous base
+	if(off > 0 && rparms.color && first) {
+		// Handle the case where the first record has ambiguous
+		// characters but we're in color space; one of those counts is
+		// spurious
+		off--;
+	}
 
 	// in now points just past the first character of a sequence
 	// line, and c holds the first character
 	while(c != -1 && c != '>') {
-		if(rparms.nsToAs && dna4Cat[c] == 2) c = 'A';
-		uint8_t cat = dna4Cat[c];
+		if(rparms.nsToAs && asc2dnacat[c] >= 2) c = 'A';
+		uint8_t cat = asc2dnacat[c];
 		int cc = toupper(c);
 		if(rparms.bisulfite && cc == 'C') c = cc = 'T';
 		if(cat == 1) {
 			// It's a DNA character
 			assert(cc == 'A' || cc == 'C' || cc == 'G' || cc == 'T');
+			// Check for overflow
+			if((TIndexOffU)(len + 1) < len) {
+				throw RefTooLongException();
+			}
 			// Consume it
 			len++;
 			// Output it
 			if(bpout != NULL) {
-				// output nucleotide
-				bpout->write(asc2dna[c]);
+				if(rparms.color) {
+					// output color
+					bpout->write(dinuc2color[asc2dna[(int)c]][lc]);
+				} else if(!rparms.color) {
+					// output nucleotide
+					bpout->write(asc2dna[c]);
+				}
 			}
-		} else if(cat == 2) {
+			lc = asc2dna[(int)c];
+		} else if(cat >= 2) {
 			// It's an N or a gap
 			lastc = c;
 			assert(cc != 'A' && cc != 'C' && cc != 'G' && cc != 'T');
@@ -140,36 +191,42 @@ RefRecord fastaRefReadSize(FileBuf& in,
 	return RefRecord((TIndexOffU)off, (TIndexOffU)len, first);
 }
 
+#if 0
 static void
 printRecords(ostream& os, const EList<RefRecord>& l) {
 	for(size_t i = 0; i < l.size(); i++) {
 		os << l[i].first << ", " << l[i].off << ", " << l[i].len << endl;
 	}
 }
+#endif
 
 /**
  * Reverse the 'src' list of RefRecords into the 'dst' list.  Don't
  * modify 'src'.
  */
-void reverseRefRecords(const EList<RefRecord>& src,
-					   EList<RefRecord>& dst,
-					   bool recursive,
-					   bool verbose)
+void reverseRefRecords(
+	const EList<RefRecord>& src,
+	EList<RefRecord>& dst,
+	bool recursive,
+	bool verbose)
 {
 	dst.clear();
 	{
 		EList<RefRecord> cur;
-		for(int64_t i = (int64_t)src.size()-1; i >= 0; i--) {
+		for(int i = (int)src.size()-1; i >= 0; i--) {
 			bool first = (i == (int)src.size()-1 || src[i+1].first);
-			if(src[i].len) {
+			// Clause after the || on next line is to deal with empty FASTA
+			// records at the end of the 'src' list, which would be wrongly
+			// omitted otherwise.
+			if(src[i].len || (first && src[i].off == 0)) {
 				cur.push_back(RefRecord(0, src[i].len, first));
 				first = false;
 			}
 			if(src[i].off) cur.push_back(RefRecord(src[i].off, 0, first));
 		}
-		for(int64_t i = 0; i < (int64_t)cur.size(); i++) {
+		for(int i = 0; i < (int)cur.size(); i++) {
 			assert(cur[i].off == 0 || cur[i].len == 0);
-			if(i < (int64_t)cur.size()-1 && cur[i].off != 0 && !cur[i+1].first) {
+			if(i < (int)cur.size()-1 && cur[i].off != 0 && !cur[i+1].first) {
 				dst.push_back(RefRecord(cur[i].off, cur[i+1].len, cur[i].first));
 				i++;
 			} else {
@@ -177,13 +234,25 @@ void reverseRefRecords(const EList<RefRecord>& src,
 			}
 		}
 	}
-	if(verbose) {
-		cout << "Source: " << endl;
-		printRecords(cout, src);
-		cout << "Dest: " << endl;
-		printRecords(cout, dst);
-	}
+	//if(verbose) {
+	//	cout << "Source: " << endl;
+	//	printRecords(cout, src);
+	//	cout << "Dest: " << endl;
+	//	printRecords(cout, dst);
+	//}
 #ifndef NDEBUG
+	size_t srcnfirst = 0, dstnfirst = 0;
+	for(size_t i = 0; i < src.size(); i++) {
+		if(src[i].first) {
+			srcnfirst++;
+		}
+	}
+	for(size_t i = 0; i < dst.size(); i++) {
+		if(dst[i].first) {
+			dstnfirst++;
+		}
+	}
+	assert_eq(srcnfirst, dstnfirst);
 	if(!recursive) {
 		EList<RefRecord> tmp;
 		reverseRefRecords(dst, tmp, true);
@@ -203,50 +272,38 @@ void reverseRefRecords(const EList<RefRecord>& src,
  * all references combined.  Rewinds each istream before returning.
  */
 std::pair<size_t, size_t>
-fastaRefReadSizes(EList<FileBuf*>& in,
-                  EList<RefRecord>& recs,
-                  EList<uint32_t>& plens,
-                  const RefReadInParams& rparms,
-                  BitpairOutFileBuf* bpout,
-                  TIndexOff& numSeqs)
+fastaRefReadSizes(
+	EList<FileBuf*>& in,
+	EList<RefRecord>& recs,
+	const RefReadInParams& rparms,
+	BitpairOutFileBuf* bpout,
+	TIndexOff& numSeqs)
 {
 	TIndexOffU unambigTot = 0;
 	size_t bothTot = 0;
 	assert_gt(in.size(), 0);
-	uint32_t both = 0, unambig = 0;
 	// For each input istream
 	for(size_t i = 0; i < in.size(); i++) {
 		bool first = true;
 		assert(!in[i]->eof());
 		// For each pattern in this istream
 		while(!in[i]->eof()) {
-			RefRecord rec = fastaRefReadSize(*in[i], rparms, first, bpout);
-			// Update plens
-			if(rec.first) {
-				if(unambig > 0) {
-					plens.push_back(both);
+			RefRecord rec;
+			try {
+				rec = fastaRefReadSize(*in[i], rparms, first, bpout);
+				if((unambigTot + rec.len) < unambigTot) {
+					throw RefTooLongException();
 				}
-				both = 0;
-				unambig = 0;
 			}
-#ifndef ACCOUNT_FOR_ALL_GAP_REFS
-			if(rec.len == 0) rec.first = false;
-#endif
-			if((unambigTot + rec.len) < unambigTot) {
-#ifdef BOWTIE_64BIT_INDEX
-				cerr << "Error: Reference sequence has more than 2^32-1 characters!  Please divide the" << endl
-				     << "reference into smaller chunks and index each independently." << endl;
-#else
-				cerr << "Error: Reference sequence has more than 2^32-1 characters!  Please try to" << endl
-				     << "build a large index instead using the appropiate options." << endl;
-#endif
+			catch(RefTooLongException& e) {
+				cerr << e.what() << endl;
 				throw 1;
 			}
 			// Add the length of this record.
 			if(rec.first) numSeqs++;
-			unambigTot += rec.len; unambig += rec.len;
-			bothTot += rec.len;    both += rec.len;
-			bothTot += rec.off;    both += rec.off;
+			unambigTot += rec.len;
+			bothTot += rec.len;
+			bothTot += rec.off;
 			first = false;
 			if(rec.len == 0 && rec.off == 0 && !rec.first) continue;
 			recs.push_back(rec);
@@ -264,9 +321,6 @@ fastaRefReadSizes(EList<FileBuf*>& in,
 	}
 	assert_geq(bothTot, 0);
 	assert_geq(unambigTot, 0);
-	if(unambig > 0) {
-		plens.push_back(both);
-	}
 	return make_pair(
 		unambigTot, // total number of unambiguous DNA characters read
 		bothTot); // total number of DNA characters read, incl. ambiguous ones
